from abc import ABC, abstractmethod
from typing import Iterable, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax import jit
from scipy.special import comb
from functools import partial
from pysersic.exceptions import * 

base_profile_types = ['sersic','doublesersic','pointsource','exp','dev']
base_profile_params =dict( 
    zip(base_profile_types,
    [ ['xc','yc','flux','r_eff','n','ellip','theta'],
    ['xc','yc','flux','f_1', 'r_eff_1','n_1','ellip_1', 'r_eff_2','n_2','ellip_2','theta'],
    ['xc','yc','flux'],
    ['xc','yc','flux','r_eff','ellip','theta'],
    ['xc','yc','flux','r_eff','ellip','theta'],]
    )
)



class BaseRenderer(object):
    def __init__(self, 
            im_shape: Iterable, 
            pixel_PSF: jax.numpy.array
            )-> None:
        """Base class for different Renderers

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        """
        self.im_shape = im_shape
        self.pixel_PSF = pixel_PSF
        if not jnp.isclose(jnp.sum(self.pixel_PSF),1.0,0.1):
            raise PSFNormalizationWarning('PSF does not appear to be appropriately normalized; Sum(psf) is more than 0.1 away from 1.')
        self.psf_shape = jnp.shape(self.pixel_PSF)
        if jnp.all(self.im_shape<self.psf_shape):
            raise KernelError('PSF pixel image size must be smaller than science image.')
        self.x = jnp.arange(self.im_shape[0])
        self.y = jnp.arange(self.im_shape[1])
        self.X,self.Y = jnp.meshgrid(self.x,self.y)
        self.x_mid = self.im_shape[0]/2. - 0.5
        self.y_mid = self.im_shape[1]/2. - 0.5

        # Set up pre-FFTed PSF
        f1d1 = jnp.fft.rfftfreq(self.im_shape[0])
        f1d2 = jnp.fft.fftfreq(self.im_shape[1])
        self.FX,self.FY = jnp.meshgrid(f1d1,f1d2)
        fft_shift_arr_x = jnp.exp(jax.lax.complex(0.,-1.)*2.*3.1415*-1*(self.psf_shape[0]/2.-0.5)*self.FX)
        fft_shift_arr_y = jnp.exp(jax.lax.complex(0.,-1.)*2.*3.1415*-1*(self.psf_shape[1]/2.-0.5)*self.FY)
        self.PSF_fft = jnp.fft.rfft2(self.pixel_PSF, s = self.im_shape)*fft_shift_arr_x*fft_shift_arr_y

        #All the  renderers here these profile types
        self.profile_types = base_profile_types
        self.profile_params = base_profile_params

    def flat_sky(self,x:float)-> float:
        """A constant sky background

        Parameters
        ----------
        x : float
            Background level
        Returns
        -------
        float
            Background level
        """
        return x

    def tilted_plane_sky(self, x:jax.numpy.array)->jax.numpy.array:
        """Render a tilted plane sky background

        Parameters
        ----------
        x : jax.numpy.array
            An array containing three parameters specifying the background

        Returns
        -------
        jax.numpy.array
            The model of the sky
        """
        return x[0] + (self.X -  self.x_mid)*x[1] + (self.Y - self.y_mid)*x[2]
    
    def render_sky(self,
            x: Union[None,float, jax.numpy.array],
            sky_type: str
            ) -> Union[float, jax.numpy.array]:
        """Render a sky background

        Parameters
        ----------
        x : Union[None,float, jax.numpy.array]
            Parameters of the given sky
        sky_type : str
            Type of sky model

        Returns
        -------
        Union[float, jax.numpy.array]
            Rendered sky background
        """
        if sky_type is None:
            return 0.
        elif sky_type == 'flat':
            return self.flat_sky(x)
        else:
            return self.tilted_plane_sky(x)

    @abstractmethod
    def render_sersic(self, xc,yc, flux, r_eff, n,ellip, theta):
        return NotImplementedError

    @abstractmethod
    def render_doublesersic(self,xc,yc, flux, f_1, r_eff_1,n_1, ellip_1, r_eff_2,n_2, ellip_2,theta):
        return NotImplementedError

    @abstractmethod
    def render_pointsource(self,xc,yc, flux):
        return NotImplementedError

    def render_exp(self, 
                xc: float,
                yc: float, 
                flux: float, 
                r_eff: float,
                ellip: float, 
                theta: float)-> jax.numpy.array:
        """Thin wrapper for an exponential profile based on render_sersic

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Exponential model
        """
        return self.render_sersic(xc,yc, flux, r_eff, 1.,ellip, theta)
    
    def render_dev(self, 
                xc: float,
                yc: float, 
                flux: float, 
                r_eff: float,
                ellip: float, 
                theta: float)-> jax.numpy.array:
        """Thin wrapper for a De Vaucouleurs profile based on render_sersic

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Exponential model
        """
        return self.render_sersic(xc,yc, flux, r_eff, 4.,ellip, theta)
    
    def render_source(self,
            params: jax.numpy.array,
            profile_type: str)->jax.numpy.array :
        """Render an observed source of a given type and parameters

        Parameters
        ----------
        params : jax.numpy.array
            Parameters specifying the source
        profile_type : str
            Type of profile to use

        Returns
        -------
        jax.numpy.array
            Rendered, observed model
        """
        render_func = getattr(self, f'render_{profile_type}')
        im = render_func(*params)
        return im


class PixelRenderer(BaseRenderer):
    """
    Render class based on rendering in pixel space and then convolving with the PSF
    """
    def __init__(self, 
            im_shape: Iterable, 
            pixel_PSF: jax.numpy.array,
            os_pixel_size: Optional[int]= 6, 
            num_os: Optional[int] = 12) -> None:
        """Initialize the PixelRenderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        os_pixel_size : Optional[int], optional
            Size of box around the center of the image to perform pixel oversampling
        num_os : Optional[int], optional
            Number of points to oversample by in each direction, by default 8
        """
        super().__init__(im_shape, pixel_PSF)
        self.os_pixel_size = os_pixel_size
        self.num_os = num_os
        
        #Use Gauss-Legendre coefficents for better integration when oversampling
        dx,w = np.polynomial.legendre.leggauss(self.num_os)
        w = w/2. 
        dx = dx/2.
        
        #dx = np.linspace(-0.5,0.5, num= num_os,endpoint=True)
        #w = np.ones_like(dx)
        self.dx_os,self.dy_os = jnp.meshgrid(dx,dx)

        w1,w2 = jnp.meshgrid(w,w)
        self.w_os = w1*w2
        
        i_mid = int(self.im_shape[0]/2)
        j_mid = int(self.im_shape[1]/2)

        self.x_os_lo, self.x_os_hi = i_mid - self.os_pixel_size, i_mid + self.os_pixel_size
        self.y_os_lo, self.y_os_hi = j_mid - self.os_pixel_size, j_mid + self.os_pixel_size

        self.X_os = self.X[self.x_os_lo:self.x_os_hi ,self.y_os_lo:self.y_os_hi ,jnp.newaxis,jnp.newaxis] + self.dx_os
        self.Y_os = self.Y[self.x_os_lo:self.x_os_hi ,self.y_os_lo:self.y_os_hi ,jnp.newaxis,jnp.newaxis] + self.dy_os


        #Set up and jit PSF convolution
        def conv(image):
            img_fft = jnp.fft.rfft2(image)
            conv_fft = img_fft*self.PSF_fft
            conv_im = jnp.fft.irfft2(conv_fft, s= self.im_shape)
            return conv_im
        self.conv = jit(conv)

        #Set up and jit intrinsic Sersic rendering with Oversampling
        def render_int_sersic(xc,yc, flux, r_eff, n,ellip, theta):
            im_no_os = render_sersic_2d(self.X,self.Y,xc,yc, flux, r_eff, n,ellip, theta)
            
            sub_im_os = render_sersic_2d(self.X_os,self.Y_os,xc,yc, flux, r_eff, n,ellip, theta)
            sub_im_os = jnp.sum(sub_im_os*self.w_os, axis = (2,3)) 
            
            im = im_no_os.at[self.x_os_lo:self.x_os_hi ,self.y_os_lo:self.y_os_hi].set(sub_im_os)
            return im
        self.render_int_sersic = jit(render_int_sersic)

    def render_sersic(self,
            xc: float,
            yc: float,
            flux: float, 
            r_eff: float, 
            n: float,
            ellip: float, 
            theta: float)->jax.numpy.array:
        """Render a sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        im_int = self.render_int_sersic(xc,yc, flux, r_eff, n,ellip, theta)
        im = self.conv(im_int)
        return im
    
    def render_doublesersic(self,
            xc: float, 
            yc: float, 
            flux: float, 
            f_1: float, 
            r_eff_1: float, 
            n_1: float, 
            ellip_1: float, 
            r_eff_2: float, 
            n_2: float, 
            ellip_2: float, 
            theta: float
            ) -> jax.numpy.array:
        """Render a double Sersic profile with a common center and position angle

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        f_1 : float
            Fraction of flux in first component
        r_eff_1 : float
            Effective radius of first component
        n_1 : float
            Sersic index of first component
        ellip_1 : float
            Ellipticity of first component
        r_eff_2 : float
            Effective radius of second component
        n_2 : float
            Sersic index of second component
        ellip_2 : float
            Ellipticity of second component
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered double Sersic model
        """
        im_int = self.render_int_sersic(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta) + self.render_int_sersic(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)
        im = self.conv(im_int)
        return im
    
    def render_pointsource(self, 
            xc: float, 
            yc: float, 
            flux: float)-> jax.numpy.array:
        """Render a Point source by interpolating given PSF into image. Currently jax only supports linear intepolation.

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        dx = xc - self.psf_shape[0]/2.
        dy = yc - self.psf_shape[1]/2.

        shifted_psf = jax.scipy.ndimage.map_coordinates(self.pixel_PSF*flux, [self.X-dx,self.Y-dy], order = 1, mode = 'constant')
    
        return shifted_psf
    
    def render_multi(self, 
            type_list: Iterable, 
            var_list: Iterable)-> jax.numpy.array:
        """Function to render multiple sources in the same image

        Parameters
        ----------
        type_list : Iterable
            List of strings containing the types of sources
        var_list : Iterable
            List of arrays contiaining the variables for each profile

        Returns
        -------
        jax.numpy.array
            Rendered image
        """
        
        int_im = jnp.zeros_like(self.X)
        obs_im = jnp.zeros_like(self.X)

        for ind in range(len(type_list)):
            if type_list[ind] == 'pointsource':
                obs_im = obs_im + self.render_pointsource(*var_list[ind])
            elif type_list[ind] == 'sersic':
                int_im = int_im + self.render_int_sersic(*var_list[ind])
            elif type_list[ind] == 'exp':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                int_im = int_im + self.render_int_sersic(xc,yc, flux, r_eff,1.,ellip, theta)
            elif type_list[ind] == 'dev':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                int_im = int_im + self.render_int_sersic(xc,yc, flux, r_eff,4.,ellip, theta)
            elif type_list[ind] == 'doublesersic':
                xc, yc, flux, f_1, r_eff_1, n_1, ellip_1, r_eff_2, n_2, ellip_2, theta = var_list[ind]
                int_im = int_im + self.render_int_sersic(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta)
                int_im = int_im + self.render_int_sersic(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)

        im = self.conv(int_im) + obs_im
        return im

class FourierRenderer(BaseRenderer):
    """
    Class to render sources based on rendering them in Fourier space. Sersic profiles are modeled as a series of Gaussian following Shajib (2019) (https://arxiv.org/abs/1906.08263) and the implementation in lenstronomy (https://github.com/lenstronomy/lenstronomy/blob/main/lenstronomy/LensModel/Profiles/gauss_decomposition.py)
    """
    def __init__(self, 
            im_shape: Iterable, 
            pixel_PSF: jax.numpy.array,
            frac_start: Optional[float] = 1e-2,
            frac_end: Optional[float] = 15., 
            n_sigma: Optional[int] = 15, 
            precision: Optional[int] = 10,
            use_poly_fit_amps: Optional[bool] = True)-> None:
        """Initialize a Fourier renderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        frac_start : Optional[float], optional
            Fraction of r_eff for the smallest Gaussian component, by default 0.01
        frac_end : Optional[float], optional
            Fraction of r_eff for the largest Gaussian component, by default 15.
        n_sigma : Optional[int], optional
            Number of Gaussian Components, by default 15
        precision : Optional[int], optional
            precision value used in calculating Gaussian components, see Shajib (2019) for more details, by default 10
        use_poly_fit_amps: Optional[bool]
            If True, instead of performing the direct calculation in Shajib (2019) at each iteration, a polynomial approximation is fit and used. The amplitudes of each gaussian component amplitudes as a function of Sersic index are fit with a polynomial. This smooth approximation is then used at each interval. While this adds a a little extra to the renderering error budget (roughly 1\%) but is much more numerically stable owing to the smooth gradients. If this matters for you then set this to False and make sure to enable jax's 64 bit capabilities which we find helps the stability.
        """
        super().__init__(im_shape, pixel_PSF)
        self.frac_start = frac_start
        self.frac_end = frac_end
        self.n_sigma = n_sigma
        self.precision = precision

        self.etas,self.betas = calculate_etas_betas(self.precision)

        if use_poly_fit_amps:

            #Set up grid of amplitudes at different values of n
            log_n_ax = jnp.logspace(jnp.log10(.65),jnp.log10(8), num = 100)
            amps_log_n = jax.vmap( lambda n: sersic_gauss_decomp(1.,1.,n,self.etas,self.betas,self.frac_start,self.frac_end,self.n_sigma)[0] ) (log_n_ax)

            #Fit polynomial for smooth interpolation
            amps_log_n_pfits = jnp.polyfit(np.log10(log_n_ax),amps_log_n,10.)

            def get_amps_sigmas(flux,r_eff,n):
                amps_norm = jnp.polyval(amps_log_n_pfits, jnp.log10(n))
                amps = amps_norm*flux
                sigmas = jnp.logspace(jnp.log10(r_eff*self.frac_start),jnp.log10(r_eff*self.frac_end),num = self.n_sigma)
                return amps,sigmas
            self.get_amps_sigmas = jax.jit(get_amps_sigmas)
        else:
            if not jax.config.jax_enable_x64:
                print ("!! WARNING !! - FourierRenderer can be numerically unstable when using jax's default 32 bit. Please either enable jax 64 bit or set 'use_poly_amps' = True in the renderer kwargs")
            def get_amps_sigmas(flux,r_eff,n):
                return sersic_gauss_decomp(flux, r_eff, n, self.etas, self.betas, self.frac_start*r_eff, self.frac_end*r_eff, self.n_sigma)
            self.get_amps_sigmas = jax.jit(get_amps_sigmas)
            

        #Jit compile function to render sersic profile in Fourier space
        def render_sersic_mog_fourier(xc,yc, flux, r_eff, n,ellip, theta):
            amps,sigmas = self.get_amps_sigmas(flux, r_eff, n)
            q = 1.-ellip
            Fgal = render_gaussian_fourier(self.FX,self.FY, amps,sigmas,xc,yc, theta,q)
            return Fgal
        self.render_sersic_mog_fourier = jit(render_sersic_mog_fourier)

        #Jit compile function to inv fft image
        def conv_and_inv_FFT(F_im):
            im = jnp.fft.irfft2(F_im*self.PSF_fft, s= self.im_shape) 
            return im
        self.conv_and_inv_FFT = jit(conv_and_inv_FFT)


    def render_sersic(self,
            xc: float,
            yc: float,
            flux: float, 
            r_eff: float, 
            n: float,
            ellip: float, 
            theta: float)->jax.numpy.array:
        """ Render a Sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        F_im = self.render_sersic_mog_fourier(xc,yc, flux, r_eff, n,ellip, theta)
        im = self.conv_and_inv_FFT(F_im)
        return im

    def render_doublesersic(self,
            xc: float, 
            yc: float, 
            flux: float, 
            f_1: float, 
            r_eff_1: float, 
            n_1: float, 
            ellip_1: float, 
            r_eff_2: float, 
            n_2: float, 
            ellip_2: float, 
            theta: float
            ) -> jax.numpy.array:
        """Render a double Sersic profile with a common center and position angle

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        f_1 : float
            Fraction of flux in first component
        r_eff_1 : float
            Effective radius of first component
        n_1 : float
            Sersic index of first component
        ellip_1 : float
            Ellipticity of first component
        r_eff_2 : float
            Effective radius of second component
        n_2 : float
            Sersic index of second component
        ellip_2 : float
            Ellipticity of second component
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered double Sersic model
        """
        F_im_1 = self.render_sersic_mog_fourier(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta)
        F_im_2 = self.render_sersic_mog_fourier(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)
        im = self.conv_and_inv_FFT(F_im_1 + F_im_2)
        return im
    
    def render_pointsource(self, 
            xc: float, 
            yc: float, 
            flux: float)-> jax.numpy.array:
        """Render a Point source

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        F_im = render_pointsource_fourier(self.FX,self.FY,xc,yc,flux)
        im = self.conv_and_inv_FFT(F_im)
        return im

    def render_multi(self, 
            type_list: Iterable, 
            var_list: Iterable)-> jax.numpy.array:
        """Function to render multiple sources in the same image

        Parameters
        ----------
        type_list : Iterable
            List of strings containing the types of sources
        var_list : Iterable
            List of arrays contiaining the variables for each profile

        Returns
        -------
        jax.numpy.array
            Rendered image
        """
        
        F_tot = jnp.zeros_like(self.FX)

        for ind in range(len(type_list)):
            if type_list[ind] == 'pointsource':
                F_tot = F_tot + render_pointsource_fourier(self.FX,self.FY,*var_list[ind])
            elif type_list[ind] == 'sersic':
                F_tot = F_tot + self.render_sersic_mog_fourier(*var_list[ind])
            elif type_list[ind] == 'exp':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                F_tot = F_tot + self.render_sersic_mog_fourier(xc,yc, flux, r_eff,1.,ellip, theta)
            elif type_list[ind] == 'dev':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                F_tot = F_tot + self.render_sersic_mog_fourier(xc,yc, flux, r_eff,4.,ellip, theta)
            elif type_list[ind] == 'doublesersic':
                xc, yc, flux, f_1, r_eff_1, n_1, ellip_1, r_eff_2, n_2, ellip_2, theta = var_list[ind]
                F_im_1 = self.render_sersic_mog_fourier(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta)
                F_im_2 = self.render_sersic_mog_fourier(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)
                F_tot = F_tot + F_im_1 + F_im_2

        im = self.conv_and_inv_FFT(F_tot)
        return im

class HybridRenderer(BaseRenderer):
    """
    Class to render sources based on the hybrid rendering scheme introduced in Lang (2020). This avoids some of the artifacts introduced by rendering sources purely in Fourier space. Sersic profiles are modeled as a series of Gaussian following Shajib (2019) (https://arxiv.org/abs/1906.08263) and the implementation in lenstronomy (https://github.com/lenstronomy/lenstronomy/blob/main/lenstronomy/LensModel/Profiles/gauss_decomposition.py).

    Our scheme is implemented slightly differently than Lang (2020), specifically in how it chooses which gaussian components to render in Fourier vs. Real space. Lang (2020) employs a cutoff based on distance to the edge of the image. However given some of jax's limitation with dynamic shapes (see more here -> https://jax.readthedocs.io/en/latest/notebooks/Common_Gotchas_in_JAX.html#dynamic-shapes), we have not implemented that specific criterion. Instead we use a simpler version where the user must decide how many components to render in real space, starting from the largest ones. While this is not ideal in all circumstances it still overcomes many of the issues of rendering purely in fourier space discussed in Lang (2020).
    """
    def __init__(self, 
            im_shape: Iterable, 
            pixel_PSF: jax.numpy.array,
            frac_start: Optional[float] = 1e-2,
            frac_end: Optional[float] = 15., 
            n_sigma: Optional[int] = 15, 
            num_pixel_render: Optional[int] = 3,
            precision: Optional[int] = 10,
            use_poly_fit_amps: Optional[bool] = True)-> None:
        """Initialize a  HybridRenderer class

        Parameters
        ----------
        im_shape : Iterable
            Tuple or list containing the shape of the desired output
        pixel_PSF : jax.numpy.array
            Pixelized version of the PSF
        frac_start : Optional[float], optional
            Fraction of r_eff for the smallest Gaussian component, by default 0.02
        frac_end : Optional[float], optional
            Fraction of r_eff for the largest Gaussian component, by default 15.
        n_sigma : Optional[int], optional
            Number of Gaussian Components, by default 13
        num_pixel_Render : Optional[int], optional
            Number of components to render in pixel space, counts back from largest component
        precision : Optional[int], optional
            precision value used in calculating Gaussian components, see Shajib (2019) for more details, by default 10
        use_poly_fit_amps: Optional[bool]
            If True, instead of performing the direct calculation in Shajib (2019) at each iteration, a polynomial approximation is fit and used. The amplitudes of each gaussian component amplitudes as a function of Sersic index are fit with a polynomial. This smooth approximation is then used at each interval. While this adds a a little extra to the renderering error budget (roughly 1\%) but is much more numerically stable owing to the smooth gradients. If this matters for you then set this to False and make sure to enable jax's 64 bit capabilities which we find helps the stability.
        """
        super().__init__(im_shape, pixel_PSF)
        self.frac_start = frac_start
        self.frac_end = frac_end
        self.n_sigma = n_sigma
        self.precision = precision
        self.num_pixel_render = num_pixel_render
        self.w_real = jnp.arange(self.n_sigma - self.num_pixel_render, self.n_sigma, dtype=jnp.int32)
        self.w_fourier = jnp.arange(self.n_sigma - self.num_pixel_render, dtype=jnp.int32)

        psf_X,psf_Y = jnp.meshgrid(jnp.arange(self.psf_shape[0]),jnp.arange(self.psf_shape[1]))
        sig_x = jnp.sqrt( (self.pixel_PSF*(psf_X)**2).sum()/self.pixel_PSF.sum() - psf_X.mean()**2 )
        sig_y = jnp.sqrt( (self.pixel_PSF*(psf_Y)**2).sum()/self.pixel_PSF.sum() - psf_Y.mean()**2 )
        self.sig_psf_approx = 0.5*(sig_x + sig_y)

        self.etas,self.betas = calculate_etas_betas(self.precision)

        if use_poly_fit_amps:

            #Set up grid of amplitudes at different values of n
            log_n_ax = jnp.logspace(jnp.log10(.65),jnp.log10(8), num = 100)
            amps_log_n = jax.vmap( lambda n: sersic_gauss_decomp(1.,1.,n,self.etas,self.betas,self.frac_start,self.frac_end,self.n_sigma)[0] ) (log_n_ax)

            #Fit polynomial for smooth interpolation
            amps_log_n_pfits = jnp.polyfit(np.log10(log_n_ax),amps_log_n,10.)

            def get_amps_sigmas(flux,r_eff,n):
                amps_norm = jnp.polyval(amps_log_n_pfits, jnp.log10(n))
                amps = amps_norm*flux
                sigmas = jnp.logspace(jnp.log10(r_eff*self.frac_start),jnp.log10(r_eff*self.frac_end),num = self.n_sigma)
                return amps,sigmas
            self.get_amps_sigmas = jax.jit(get_amps_sigmas)
        else:
            if not jax.config.jax_enable_x64:
                print ("!! WARNING !! - HybridRenderer can be numerically unstable when using jax's default 32 bit. Please either enable jax 64 bit or set 'use_poly_amps' = True in the renderer kwargs")
            def get_amps_sigmas(flux,r_eff,n):
                return sersic_gauss_decomp(flux, r_eff, n, self.etas, self.betas, self.frac_start*r_eff, self.frac_end*r_eff, self.n_sigma)
            self.get_amps_sigmas = jax.jit(get_amps_sigmas)
            

        def render_sersic_hybrid(xc,yc, flux, r_eff, n,ellip, theta):
            amps,sigmas = self.get_amps_sigmas(flux, r_eff, n)
            
            q = 1.-ellip

            sigmas_obs = jnp.sqrt(sigmas**2 + self.sig_psf_approx**2)
            q_obs = jnp.sqrt( (q*q*sigmas**2 + self.sig_psf_approx**2)/ sigmas_obs**2 )

            Fgal = render_gaussian_fourier(self.FX,self.FY, amps[self.w_fourier],sigmas[self.w_fourier],xc,yc, theta,q)

            im_gal = render_gaussian_pixel(self.X,self.Y, amps[self.w_real],sigmas_obs[self.w_real],xc,yc, theta,q_obs[self.w_real])

            return Fgal,im_gal
        self.render_sersic_hyrbid = jax.jit(render_sersic_hybrid)

        def conv_and_inv_FFT(F_im):
            im = jnp.fft.irfft2(F_im*self.PSF_fft, s= self.im_shape) 
            return im
        self.conv_and_inv_FFT = jit(conv_and_inv_FFT)

    def render_sersic(self,
            xc: float,
            yc: float,
            flux: float, 
            r_eff: float, 
            n: float,
            ellip: float, 
            theta: float)->jax.numpy.array:
        """ Render a Sersic profile

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        r_eff : float
            Effective radius
        n : float
            Sersic index
        ellip : float
            Ellipticity
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered Sersic model
        """
        F, im  = self.render_sersic_hyrbid(xc,yc, flux, r_eff, n,ellip, theta)
        out = im + self.conv_and_inv_FFT(F)
        return out

    def render_doublesersic(self,
            xc: float, 
            yc: float, 
            flux: float, 
            f_1: float, 
            r_eff_1: float, 
            n_1: float, 
            ellip_1: float, 
            r_eff_2: float, 
            n_2: float, 
            ellip_2: float, 
            theta: float
            ) -> jax.numpy.array:
        """Render a double Sersic profile with a common center and position angle

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux
        f_1 : float
            Fraction of flux in first component
        r_eff_1 : float
            Effective radius of first component
        n_1 : float
            Sersic index of first component
        ellip_1 : float
            Ellipticity of first component
        r_eff_2 : float
            Effective radius of second component
        n_2 : float
            Sersic index of second component
        ellip_2 : float
            Ellipticity of second component
        theta : float
            Position angle in radians

        Returns
        -------
        jax.numpy.array
            Rendered double Sersic model
        """
        F_1, im_1 = self.render_sersic_hyrbid(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta)
        F_2, im_2 = self.render_sersic_hyrbid(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)
        im = im_1 + im_2  + self.conv_and_inv_FFT(F_1 + F_2)
        return im

    def render_pointsource(self, 
            xc: float, 
            yc: float, 
            flux: float)-> jax.numpy.array:
        """Render a Point source

        Parameters
        ----------
        xc : float
            Central x position
        yc : float
            Central y position
        flux : float
            Total flux

        Returns
        -------
        jax.numpy.array
            rendered pointsource model
        """
        F_im = render_pointsource_fourier(self.FX,self.FY,xc,yc,flux)
        im = self.conv_and_inv_FFT(F_im)
        return im


    def render_multi(self, 
            type_list: Iterable, 
            var_list: Iterable)-> jax.numpy.array:
        """Function to render multiple sources in the same image

        Parameters
        ----------
        type_list : Iterable
            List of strings containing the types of sources
        var_list : Iterable
            List of arrays contiaining the variables for each profile

        Returns
        -------
        jax.numpy.array
            Rendered image
        """
        
        F_tot = jnp.zeros_like(self.FX)
        im_tot = jnp.zeros_like(self.X)

        for ind in range(len(type_list)):
            if type_list[ind] == 'pointsource':
                F_tot = F_tot + render_pointsource_fourier(self.FX,self.FY,*var_list[ind])
            elif type_list[ind] == 'sersic':
                F_cur, im_cur  = self.render_sersic_hyrbid(*var_list[ind])
                F_tot = F_tot + F_cur
                im_tot = im_tot + im_cur
            elif type_list[ind] == 'exp':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                F_cur, im_cur  = self.render_sersic_hyrbid(xc,yc, flux, r_eff,1.,ellip, theta)
                F_tot = F_tot + F_cur
                im_tot = im_tot + im_cur
            elif type_list[ind] == 'dev':
                xc,yc, flux, r_eff,ellip, theta = var_list[ind]
                F_cur, im_cur  = self.render_sersic_hyrbid(xc,yc, flux, r_eff,4.,ellip, theta)
                F_tot = F_tot + F_cur
            elif type_list[ind] == 'doublesersic':
                xc, yc, flux, f_1, r_eff_1, n_1, ellip_1, r_eff_2, n_2, ellip_2, theta = var_list[ind]
                F_1, im_1 = self.render_sersic_hyrbid(xc,yc, flux*f_1, r_eff_1, n_1,ellip_1, theta)
                F_2, im_2 = self.render_sersic_hyrbid(xc,yc, flux*(1.-f_1), r_eff_2, n_2,ellip_2, theta)
                F_tot = F_tot + F_1 + F_2
                im_tot = im_tot + im_1 + im_2

        im = self.conv_and_inv_FFT(F_tot) + im_tot
        return im

@jax.jit
def sersic1D(
        r: Union[float, jax.numpy.array],
        flux: float,
        re: float,
        n: float)-> Union[float, jax.numpy.array]:
    """Evaluate a 1D sersic profile

    Parameters
    ----------
    r : float
        radii to evaluate profile at
    flux : float
        Total flux
    re : float
        Effective radius
    n : float
        Sersic index

    Returns
    -------
    jax.numpy.array
        Sersic profile evaluated at r
    """
    bn = 1.9992*n - 0.3271
    Ie = flux / ( re*re* 2* jnp.pi*n * jnp.exp(bn + jax.scipy.special.gammaln(2*n) ) ) * bn**(2*n) 
    return Ie*jnp.exp ( -bn*( (r/re)**(1./n) - 1. ) )


@jax.jit
def render_gaussian_fourier(FX: jax.numpy.array,
        FY: jax.numpy.array,
        amps: jax.numpy.array,
        sigmas: jax.numpy.array,
        xc: float,
        yc: float, 
        theta: float,
        q: float)-> jax.numpy.array:
    """Render Gaussian components in the Fourier domain

    Parameters
    ----------
    FX : jax.numpy.array
        X frequency positions to evaluate
    FY : jax.numpy.array
        Y frequency positions to evaluate
    amps : jax.numpy.array
        Amplitudes of each component
    sigmas : jax.numpy.array
        widths of each component
    xc : float
        Central x position
    yc : float
        Central y position
    theta : float
        position angle
    q : float
        Axis ratio

    Returns
    -------
    jax.numpy.array
        Sum of components evaluated at FX and FY
    """
    Ui = FX*jnp.cos(theta) + FY*jnp.sin(theta) 
    Vi = -1*FX*jnp.sin(theta) + FY*jnp.cos(theta) 

    in_exp = -1*(Ui*Ui + Vi*Vi*q*q)*(2*jnp.pi*jnp.pi*sigmas*sigmas)[:,jnp.newaxis,jnp.newaxis] - 1j*2*jnp.pi*FX*xc - 1j*2*jnp.pi*FY*yc
    Fgal_comp = amps[:,jnp.newaxis,jnp.newaxis]*jnp.exp(in_exp)
    Fgal = jnp.sum(Fgal_comp, axis = 0)
    return Fgal

@jax.jit
def render_pointsource_fourier(FX: jax.numpy.array,
        FY: jax.numpy.array,
        xc: float,
        yc: float, 
        flux: float)-> jax.numpy.array:
    """Render a point source in the Fourier domain

    Parameters
    ----------
    FX : jax.numpy.array
        X frequency positions to evaluate
    FY : jax.numpy.array
        Y frequency positions to evaluate
    xc : float
        Central x position
    yc : float
        Central y position
    flux : float
        Total flux of source

    Returns
    -------
    jax.numpy.array
        Point source evaluated at FX FY
    """
    in_exp = -1j*2*jnp.pi*FX*xc - 1j*2*jnp.pi*FY*yc
    F_im = flux*jnp.exp(in_exp)
    return F_im


@jax.jit
def render_gaussian_pixel(X: jax.numpy.array,
        Y: jax.numpy.array,
        amps: jax.numpy.array,
        sigmas: jax.numpy.array,
        xc: float,
        yc: float, 
        theta: float,
        q: Union[float,jax.numpy.array])-> jax.numpy.array:
    """Render Gaussian components in pixel space

    Parameters
    ----------
    FX : jax.numpy.array
        X positions to evaluate
    FY : jax.numpy.array
        Y positions to evaluate
    amps : jax.numpy.array
        Amplitudes of each component
    sigmas : jax.numpy.array
        widths of each component
    xc : float
        Central x position
    yc : float
        Central y position
    theta : float
        position angle
    q : Union[float,jax.numpy.array]
        Axis ratio

    Returns
    -------
    jax.numpy.array
        Sum of components evaluated at X and Y
    """
    X_bar = X - xc
    Y_bar = Y - yc

    Xi = X_bar*jnp.cos(theta) + Y_bar*jnp.sin(theta) 
    Yi = -1*X_bar*jnp.sin(theta) + Y_bar*jnp.cos(theta) 

    in_exp = -1*(Xi*Xi + Yi*Yi/(q*q)[:,jnp.newaxis,jnp.newaxis] )/ (2*sigmas*sigmas)[:,jnp.newaxis,jnp.newaxis]
    im_comp = (amps/(2*jnp.pi*sigmas*sigmas*q))[:,jnp.newaxis,jnp.newaxis]*jnp.exp(in_exp)
    im = jnp.sum(im_comp, axis = 0)
    return im


@jax.jit
def render_sersic_2d(X: jax.numpy.array,
    Y: jax.numpy.array,
    xc: float,
    yc: float, 
    flux: float, 
    r_eff: float,
    n: float,
    ellip: float, 
    theta: float)-> jax.numpy.array:
    """Evalulate a 2D Sersic distribution at given locations

    Parameters
    ----------
    X : jax.numpy.array
        x locations to evaluate at
    Y : jax.numpy.array
        y locations to evaluate at
    xc : float
        Central x position
    yc : float
        Central y position
    flux : float
        Total flux
    r_eff : float
        Effective radius
    n : float
        Sersic index
    ellip : float
        Ellipticity
    theta : float
        Position angle in radians


    Returns
    -------
    jax.numpy.array
        Sersic model evaluated at given locations
    """
    bn = 1.9992*n - 0.3271
    a, b = r_eff, (1 - ellip) * r_eff
    cos_theta, sin_theta = jnp.cos(theta), jnp.sin(theta)
    x_maj = (X - xc) * cos_theta + (Y - yc) * sin_theta
    x_min = -(X - xc) * sin_theta + (Y - yc) * cos_theta
    amplitude = flux*bn**(2*n) / ( jnp.exp(bn + jax.scipy.special.gammaln(2*n) ) *r_eff**2 *jnp.pi*2*n )
    z = jnp.sqrt((x_maj / a) ** 2 + (x_min / b) ** 2)
    out = amplitude * jnp.exp(-bn * (z ** (1 / n) - 1)) / (1.-ellip)
    return out

def calculate_etas_betas(precision: int)-> Tuple[jax.numpy.array, jax.numpy.array]:
    """Calculate the weights and nodes for the Gaussian decomposition described in Shajib (2019) (https://arxiv.org/abs/1906.08263)

    Parameters
    ----------
    precision : int
        Precision, higher number implies more precise decomposition but more nodes. Effective upper limit is 12 for 32 bit numbers, 27 for 64 bit numbers.

    Returns
    -------
    Tuple[jax.numpy.array, jax.numpy.array]
        etas and betas array to be use in gaussian decomposition
    """
    kes = jnp.arange(2 * precision + 1)
    betas = jnp.sqrt(2 * precision * jnp.log(10) / 3. + 2. * 1j * jnp.pi * kes)
    epsilons = jnp.zeros(2 * precision + 1)

    epsilons = epsilons.at[0].set(0.5)
    epsilons = epsilons.at[1:precision + 1].set(1.)
    epsilons = epsilons.at[-1].set(1 / 2. ** precision)

    for k in range(1, precision):
        epsilons = epsilons.at[2 * precision - k].set(epsilons[2 * precision - k + 1] + 1 / 2. ** precision * comb(precision, k) )

    etas = jnp.array( (-1.) ** kes * epsilons * 10. ** (precision / 3.) * 2. * jnp.sqrt(2*jnp.pi) )
    betas = jnp.array(betas)
    return etas,betas

def sersic_gauss_decomp(
        flux: float, 
        re: float, 
        n:float, 
        etas: jax.numpy.array, 
        betas:jax.numpy.array, 
        sigma_start: float, 
        sigma_end: float, 
        n_comp: int)-> Tuple[jax.numpy.array, jax.numpy.array]:
    """Calculate a gaussian decomposition of a given sersic profile, following Shajib (2019) (https://arxiv.org/abs/1906.08263)

    Parameters
    ----------
    flux : float
        Total flux
    re : float
        half light radius
    n : float
        Sersic index
    etas : jax.numpy.array
        Weights for decomposition, can be calcualted using pysersic.rendering_utils.calculate_etas_betas
    betas : jax.numpy.array
        Nodes for decomposition, can be calcualted using pysersic.rendering_utils.calculate_etas_betas
    sigma_start : float
        width for the smallest Gaussian component
    sigma_end : float
         width for the largest Gaussian component
    n_comp : int
        Number of Gaussian components

    Returns
    -------
    Tuple[jax.numpy.array, jax.numpy.array]
        Amplitudes and sigmas of Gaussian decomposition
    """
    sigmas = jnp.logspace(jnp.log10(sigma_start),jnp.log10(sigma_end),num = n_comp)

    f_sigmas = jnp.sum( etas * sersic1D(jnp.outer(sigmas,betas),flux,re,n).real,  axis=1)

    del_log_sigma = jnp.abs(jnp.diff(jnp.log(sigmas)).mean())

    amps = f_sigmas * del_log_sigma / jnp.sqrt(2*jnp.pi)

    amps = amps.at[0].multiply(0.5)
    amps = amps.at[-1].multiply(0.5)

    amps = amps*2*jnp.pi*sigmas*sigmas

    return amps,sigmas



from __future__ import annotations
from abc import ABC
from typing import Iterable, Optional, Union, Tuple

import astropy.units as u
from astropy.stats import biweight_scale as bws
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import pandas
from numpyro import distributions as dist, infer, sample
from photutils.morphology import data_properties

base_sky_types = ['none','flat','tilted-plane']
base_sky_params = dict(
    zip(base_sky_types,
    [ [],['sky_back',], ['sky_back','sky_x_sl','sky_y_sl'] ]
    )
)

@jax.jit
def render_tilted_plane_sky(X,Y,back,x_sl,y_sl ):
    xmid = float(X.shape[0]/2.)
    ymid = float(Y.shape[0]/2.)
    return back + (X-xmid)*x_sl + (Y-ymid)*y_sl


base_profile_types = ['sersic','doublesersic','pointsource','exp','dev']
base_profile_params =dict( 
    zip(base_profile_types,
    [ ['xc','yc','flux','r_eff','n','ellip','theta'],
    ['xc','yc','flux','f_1', 'r_eff_1','n_1','ellip_1', 'r_eff_2','n_2','ellip_2','theta'],
    ['xc','yc','flux'],
    ['xc','yc','flux','r_eff','ellip','theta'],
    ['xc','yc','flux','r_eff','ellip','theta'],]
    )
)


class BasePrior(ABC):
    """
    Base class for priors with sky sampling included
    """
    def __init__(self, sky_type = 'none',sky_guess=None,sky_guess_err=None) -> None:
        """Initialize a base prior class

        Parameters
        ----------
        sky_type : str, optional
            Type of sky modle to use, one of: none,flat or tilted-plane, by default 'none'
        """
        self.reparam_dict = {}
        self.sky_type = sky_type
        if sky_guess is None:
            self.sky_guess = 0.0 
        else:
            assert sky_guess is not None and sky_guess_err is not None, 'If using fitting a sky then must supply a guess and uncertainty on the background value'
            self.sky_guess = sky_guess
            self.sky_guess_err = sky_guess_err

        if self.sky_type not in base_sky_types:
            raise AssertionError("Sky type must be one of: ", base_sky_types)
        elif self.sky_type == 'none':
            self.sample_sky = self.sample_sky_none
    
        elif self.sky_type == 'flat':
            
            self._set_dist('sky_back',dist.TransformedDistribution(
                                dist.Normal(),
                                dist.transforms.AffineTransform(sky_guess,sky_guess_err),)
                            )
            self.reparam_dict['sky_back'] = infer.reparam.TransformReparam()
            self.sample_sky = self.sample_sky_flat
            
        elif self.sky_type == 'tilted-plane':
            self._set_dist('sky_back',dist.TransformedDistribution(
                                dist.Normal(),
                                dist.transforms.AffineTransform(sky_guess,sky_guess_err),)
                            )
            self._set_dist('sky_x_sl',  dist.TransformedDistribution(
                                dist.Normal(),
                                dist.transforms.AffineTransform(0.0,0.1*sky_guess_err),)
            )

            self._set_dist('sky_y_sl',dist.TransformedDistribution(
                                dist.Normal(),
                                dist.transforms.AffineTransform(0.0,0.1*sky_guess_err),)
            )
            
            self.reparam_dict['sky_back'] = infer.reparam.TransformReparam()
            self.reparam_dict['sky_x_sl'] = infer.reparam.TransformReparam()
            self.reparam_dict['sky_y_sl'] = infer.reparam.TransformReparam()
            self.sample_sky = self.sample_sky_tilted_plane

    def sample_sky_none(self,X: jax.numpy.array,Y: jax.numpy.array)-> float:
        """Simple wrapper to generate no sky

        Parameters
        ----------
        X : jax.numpy.array
            2d array, X pixel values
        Y : jax.numpy.array
            2d array, Y pixel values

        Returns
        -------
        float = 0
            returns 0
        """
        return 0.
    
    def sample_sky_flat(self,X: jax.numpy.array,Y: jax.numpy.array)-> float:
        """
        Sample and generate a flat sky background

        Parameters
        ----------
        X : jax.numpy.array
            2d array, X pixel values
        Y : jax.numpy.array
            2d array, Y pixel values

        Returns
        -------
        float
            sampled background value
        """
        sky_back = sample('sky_back', self.sky_back_prior_dist)
        return sky_back
    
    def sample_sky_tilted_plane(self,X: jax.numpy.array,Y: jax.numpy.array)-> jax.numpy.array:
        """
        Sample and generate a tilted-plane sky background

        Parameters
        ----------
        X : jax.numpy.array
            2d array, X pixel values
        Y : jax.numpy.array
            2d array, Y pixel values

        Returns
        -------
        jax.numpy.array
            renderned sky background
        """
        sky_back = sample('sky_back', self.sky_back_prior_dist)
        sky_x_sl = sample('sky_x_sl', self.sky_x_sl_prior_dist)
        sky_y_sl = sample('sky_y_sl', self.sky_y_sl_prior_dist)
        return render_tilted_plane_sky(X,Y, sky_back,sky_x_sl,sky_y_sl)

    def _set_dist(self, var_name: str, dist: dist.Distribution)-> None:
        """Set prior for a given variable

        Parameters
        ----------
        var_name : str
            variable name
        dist : dist.Distribution
            Numpyro distribution object specifying prior
        """
        self.__setattr__(var_name+'_prior_dist', dist)

    def _get_dist(self, var_name: str)-> dist.Distribution:
        """
        Get prior for a given variable

        Parameters
        ----------
        var_name : str
            variable name

        Returns
        -------
        dist.Distribution
            Numpyro distribution clas describing prior
        """
        return self.__getattribute__(var_name +'_prior_dist')
    

class PySersicSourcePrior(BasePrior):
    """
    Class used for priors for single source fitting in PySersic
    """
    def __init__(self, 
                profile_type: str, 
                sky_type: Optional[str] = 'none',
                sky_guess: Optional[float]=None,
                sky_guess_err: Optional[float] = None,
                suffix: Optional[str] =  "") -> None:
        """Initialize PySersicSourcePrior class

        Parameters
        ----------
        profile_type : str
            Type of profile
        sky_type : Optional[str], optional
            Type of sky model to use, one of: none, flat, tilted-plane, by default 'none'
        suffix : Optional[str], optional
            Additional suffix to add to each variable name, used in PySersicMultiPrior, by default ""
        """
        super().__init__(sky_type= sky_type,sky_guess=sky_guess,sky_guess_err=sky_guess_err)
        assert profile_type in base_profile_types
        self.profile_type = profile_type
        self.param_names = base_profile_params[self.profile_type]
        self.repr_dict = {}
        self.built = False
        self.suffix = suffix

    def __repr__(self, print_sky = True) -> str:
        out = f"Prior for a {self.profile_type} source:"
        num_dash = len(out)
        out += "\n" + "-"*num_dash + "\n"
        for (var, descrip) in self.repr_dict.items():
            out += var + " ---  " + descrip + "\n"    
        if print_sky: # This is not as dynamic as the parameter ones so will not update once set
            out+= f'\nSky Type: {self.sky_type}\n'
            if not self.sky_type == 'none':
                out+=f'Sky level --- Normal w/ mu = {self.sky_guess:.1e}, sigma = {self.sky_guess_err:.1e}\n'
                if self.sky_type == 'tilted-plane':
                    out+=f'Sky x slope --- Normal w/ mu = {0}, sigma = {self.sky_guess_err*0.1:.1e}\n'
                    out+=f'Sky y slope --- Normal w/ mu = {0}, sigma = {self.sky_guess_err*0.1:.1e}\n'
        return out
    
    def _build_dist_list(self)-> None:
        """
        Function to combine all distributions into list, sets self.dist_list
        """
        self.dist_list = []
        assert self.check_vars() # check and make sure all is good

        for param in self.param_names:
            self.dist_list.append(self._get_dist(param+self.suffix))
        
        self.built = True
    
    def __call__(self) -> jax.numpy.array:
        """
        Sample variables from prior

        Returns
        -------
        jax.numpy.array
            sampled variables
        """
        if not self.built: 
            self._build_dist_list()

        arr = []
        for (param,prior) in zip(self.param_names,self.dist_list):
            if issubclass(type(prior), dist.Distribution):
                arr.append(sample(param+self.suffix, prior) )
            else:
                arr.append(prior)
        return jnp.array(arr)      

    def set_gaussian_prior(self,var_name: str, loc: float, scale: float) -> "PySersicSourcePrior":
        """
        Set a Gaussian prior for a variable

        Parameters
        ----------
        var_name : str
            variable name
        loc : float
            mean 
        scale : float
            standard deviation

        Returns
        -------
        PySersicSourcePrior
            returns self to allow chaining
        """

        prior_dist = dist.TransformedDistribution(
            dist.Normal(),
            dist.transforms.AffineTransform(loc,scale),)
        
        self._set_dist(var_name + self.suffix, prior_dist)
        self.reparam_dict[var_name  + self.suffix] = infer.reparam.TransformReparam()
        self.repr_dict[var_name] = f"Normal w/ mu = {loc:.2f}, sigma = {scale:.2f}"
        return self
    
    def set_uniform_prior(self, var_name:str, low: float,high: float)-> "PySersicSourcePrior":
        """
        Set a uniform prior for a variable

        Parameters
        ----------
        var_name : str
            variable name
        low : float
            lower bound
        high : float
            upper bound

        Returns
        -------
        PySersicSourcePrior
            returns self to allow chaining

        """
        shift = low
        scale = high-low
        prior_dist = dist.TransformedDistribution(
            dist.Uniform(),
            dist.transforms.AffineTransform(shift,scale),)
        self._set_dist(var_name + self.suffix, prior_dist)
        self.reparam_dict[var_name + self.suffix] = infer.reparam.TransformReparam()
        self.repr_dict[var_name] = f"Uniform between: {low:.2f} -> {high:.2f}"
        return self
    
    def set_truncated_gaussian_prior(self,
            var_name:str, 
            loc: float,
            scale:float, 
            low: Optional[float] = None,
            high: Optional[float] = None) -> "PySersicSourcePrior":
        """
        Set a truncated Gaussian prior for a given variable

        Parameters
        ----------
        var_name : str
            variable name
        loc : float
            mean
        scale : float
            standard deviation
        low : Optional[float], optional
            lower bound, by default None
        high : Optional[float], optional
            upper bound, by default None

        Returns
        -------
        PySersicSourcePrior
            Returns self to allow chaining
        """
        if low is not None:
            low_scaled = (low - loc)/scale
        else:
            low = -jnp.inf
            low_scaled = None
        
        if high is not None:
            high_scaled = (high - loc)/scale
        else:
            high_scaled = None
            high = jnp.inf
        prior_dist = dist.TransformedDistribution(
            dist.TruncatedNormal(low= low_scaled, high = high_scaled),
            dist.transforms.AffineTransform(loc,scale),)
    
        self._set_dist(var_name + self.suffix, prior_dist)
        self.reparam_dict[var_name + self.suffix] = infer.reparam.TransformReparam()
        self.repr_dict[var_name] = f"Truncated Normal w/ mu = {loc:.2f}, sigma = {scale:.2f}, between: {low:.2f} -> {high:.2f}"
        return self

    def set_custom_prior(self, 
            var_name: str, 
            prior_dist: dist.Distribution, 
            reparam: Optional[infer.reparam.Reparam] = None)-> "PySersicSourcePrior":
        """Set a custom distribution as the prior for a given variable

        Parameters
        ----------
        var_name : str
            variable name
        prior_dist : dist.Distribution
            Numpyro Distribution object describing prior
        reparam : Optional[infer.reparam.Reparam], optional
            Optional reparamaterization to use for variable, by default None

        Returns
        -------
        PySersicSourcePrior
            Returns self to allow chaining
        """
        self._set_dist(var_name + self.suffix, prior_dist)
        if reparam is not None:
            self.reparam_dict[var_name + self.suffix] = reparam
        self.repr_dict[var_name] = "Custom prior of type: "+ str(prior_dist.__class__)
        return self
    
    def check_vars(self, verbose = False) -> bool:
        """
        Function to check if all variable for the specified profile type are set with no extras

        Parameters
        ----------
        verbose : bool, optional
            Wheter to print out missing and extra variables, by default False

        Returns
        -------
        bool
            True if all variable for given type are present with no extra, False otherwise
        """
        missing = []
        for var in self.param_names:
            if not hasattr(self, f"{var + self.suffix}_prior_dist"):
                missing.append(var)
        
        extra = []
        for (name, descrip) in self.repr_dict.items():
            if (name not in self.param_names) and ('sky' not in name):
                extra.append(name)
        if verbose:
            print ("Missing params for profile: ", missing)
            print ("Extra params, will not be used: ", extra)
        
        if len(missing)== 0 and len(extra)==0:
            out = True
        else:
            out = False
        return out
    
        
class PySersicMultiPrior(BasePrior):
    """
    Class used for priors for multi source fitting in PySersic
    """
    def __init__(self, 
            catalog: Union[pandas.DataFrame,dict, np.recarray],
            sky_type: Optional[str] = 'none', 
            sky_guess: Optional[float] = None,
            sky_guess_err: Optional[float] = None    
            )-> None:
        """
        Ingest a catalog-like data structure containing prior positions and parameters for multiple sources in a single image. 
        The format of the catalog can be a `pandas.DataFrame`, `numpy` RecordArray, dictionary, or any other format so-long as 
        the following fields exist and can be directly indexed: 'x', 'y', 'flux', 'r' and 'type'

        Parameters
        ----------
        image : jax.numpy.array
            science image
        catalog : Union[pandas.DataFrame,dict, np.recarray]
            Object containing information about the sources to be fit
        Returns
        -------
        prior_list : Iterable
            List containing a prior dictionary for each source
        """
        image = jnp.zeros((100,100)) # dummy image
        properties = SourceProperties(image)
        if sky_type != 'none':
                assert sky_guess is not None and sky_guess_err is not None, "If using a sky model must provide initial guess and uncertainty"

        super().__init__(sky_type = sky_type,sky_guess=sky_guess,sky_guess_err=sky_guess_err)
        
        self.catalog = catalog
        self.all_priors = []
        self.N_sources = len(catalog['x'])
        image = jnp.zeros((100,100)) # dummy image
        
        #Loop through catalog to generate priors
        for ind in range(len(catalog['x'])):
            properties.set_flux_guess(catalog['flux'][ind])
            properties.set_r_eff_guess(r_eff_guess = catalog['r'][ind])
            properties.set_position_guess((catalog['x'][ind],catalog['y'][ind]) )
            
            try:
                properties.set_theta_guess(catalog['theta'][ind])
            except:
                pass 
            
            prior = properties.generate_prior(catalog['type'][ind], sky_type= 'none', suffix  = f'_{ind:d}')

            self.all_priors.append(prior)
            self.reparam_dict.update(prior.reparam_dict)
    
    def __repr__(self,)-> str:
        out = f"PySersicMultiPrior containing {len(self.all_priors):d} sources \n"
        for i, source_prior in enumerate(self.all_priors):
            out += f"\nSource {i:d} : " + source_prior.__repr__(print_sky = False)
        out+= f'\nSky Type: {self.sky_type}\n'
        if not self.sky_type == 'none':
            out+=f'Sky level --- Normal w/ mu = {self.sky_guess:.1e}, sigma = {self.sky_guess_err:.1e}\n'
            if self.sky_type == 'tilted-plane':
                out+=f'Sky x slope --- Normal w/ mu = {0}, sigma = {self.sky_guess_err*0.1:.1e}\n'
                out+=f'Sky y slope --- Normal w/ mu = {0}, sigma = {self.sky_guess_err*0.1:.1e}\n'
        return out
    
    def __call__(self) -> list:
        """Sample prior for all sources

        Returns
        -------
        list
            a list of jax.numpy.arrays containing sampled variables for each source
        """
        all_params = []
        for prior_cur in self.all_priors:
            all_params.append(prior_cur())
        return all_params



class SourceProperties():
    """
    A Class used to estimate initial guesses for source properties. If no guesses are provided, then the class will estimate them using the `photutls` package and the `data_properties()` function.
    """
    def __init__(self,image: np.array, mask:np.array =None):
        """Initialize the a SourceProperties object

        Parameters
        ----------
        image : np.array
            science image
        mask : np.array, optional
            pixel by pixel mask, by default None
        """
        self.image = image 
        self.mask = mask 
        
        if self.mask is not None:
            self.cat = data_properties(self.image,mask=self.mask.astype(bool))
            self.image = np.ma.masked_array(self.image, self.mask )
        else:
            self.cat = data_properties(self.image)
        _ = self.measure_properties() 

    def measure_properties(self,**kwargs) -> SourceProperties:
        """Measure default properties of the source

        Returns
        -------
        SourceProperties
            returns self
        """
        self.set_flux_guess(**kwargs)
        self.set_r_eff_guess(**kwargs)
        self.set_theta_guess(**kwargs)
        self.set_position_guess(**kwargs)
        self.set_sky_guess(**kwargs)
        return self
    
    def set_sky_guess(self,sky_guess: Optional[float] = None,sky_guess_err:Optional[float] = None,n_pix_sample:int = 5, **kwargs)-> SourceProperties:
        """Measure or set guess for initial sky background level. If no estimate is provided, the median of the n_pix_sample number of pixels around each edge is used

        Parameters
        ----------
        sky_guess : Optional[float], optional
            Initial guess for level of background, by default None
        sky_guess_err : Optional[float], optional
            Uncertainity on inital guess, by default None
        n_pix_sample : int, optional
            Number of pixels around each edge to use to estimate sky level if neccesary, by default 5

        Returns
        -------
        SourceProperties
            returns self
        """
        med, std, npix = estimate_sky(self.image, n_pix_sample= n_pix_sample)
        if sky_guess is None:
            self.sky_guess = med
        if sky_guess_err is None:
            self.sky_guess_err = 2*std/np.sqrt(npix)
        return self

    def set_flux_guess(self,flux_guess: Optional[float] =None,flux_guess_err: Optional[float] = None,**kwargs)-> SourceProperties:
        """Measure or set guess for initial flux. If no estimate is provided, the flux of the source in estimated as the total flux within the sgmentatated region for the source

        Parameters
        ----------
        flux_guess : Optional[float], optional
            Initial guess for flux, by default None
        flux_guess_err : Optional[float], optional
            Uncertainty on initial guess, by default None

        Returns
        -------
        SourceProperties
            returns self
        """
        if flux_guess is None:
            flux_guess = self.cat.segment_flux
        if flux_guess_err is not None:
            flux_guess_err = flux_guess_err
        else:
            if flux_guess > 0:
                flux_guess_err = 2*np.sqrt( flux_guess )
            else:
                flux_guess_err = 2*np.sqrt(np.abs(flux_guess))
                flux_guess = 0.
        
        self.flux_guess = flux_guess 
        self.flux_guess_err = flux_guess_err
        return self 
    
    def set_r_eff_guess(self,r_eff_guess:Optional[float] = None, r_eff_guess_err:Optional[float] = None, **kwargs)-> SourceProperties:
        """Measure or set guess for effective radius. If no estimate is provided, the r_eff of the source in estimated using photutils

        Parameters
        ----------
        r_eff_guess : Optional[float], optional
            Initial guess for effective radius, by default None
        r_eff_guess_err : Optional[float], optional
            Uncertainty on initial guess, by default None

        Returns
        -------
        SourceProperties
            returns self
        """
        if r_eff_guess is None:
            r_eff_guess = self.cat.fluxfrac_radius(0.5).to(u.pixel).value
        
        if r_eff_guess_err is not None:
            self.r_eff_guess_err = r_eff_guess_err
        else:
            self.r_eff_guess_err = 2*np.sqrt(r_eff_guess) 

        self.r_eff_guess = r_eff_guess
        return self

    def set_theta_guess(self,theta_guess: Optional[float] = None,**kwargs)-> SourceProperties:
        """Measure or set guess for initial position angle. If no estimate is provided, the position angle of the source in estimated using the data_properties() function from photutils

        Parameters
        ----------
        theta_guess : Optional[float], optional
            Estimate of the position angle in radians, by default None

        Returns
        -------
        SourceProperties
            returns self
        """

        if theta_guess is None:
            theta_guess = self.cat.orientation.to(u.rad).value

        if np.isnan(theta_guess):
            theta_guess = 0
        self.theta_guess = theta_guess 
        return self
    
    def set_position_guess(self,position_guess: Optional[Iterable[float,float]]=None,**kwargs)-> SourceProperties:
        """Measure or set guess for initial position. If no estimate is provided, the position of the source in estimated using the data_properties() function from photutils

        Parameters
        ----------
        position_guess : Optional[Iterable[float,float]], optional
            A 2 element list, tuple or array which contain the x,y pixel values of the inital guess for the centroid, by default None

        Returns
        -------
        SourceProperties
            returns self
        """
        if position_guess is None:
            self.xc_guess = self.cat.centroid_win[0]
            self.yc_guess = self.cat.centroid_win[1]
        else:
            self.xc_guess = position_guess[0]
            self.yc_guess = position_guess[1]
        return self
    
    def generate_prior(self,
                profile_type: str,
                sky_type: Optional[str] = 'none',
                suffix: Optional[str] = '')-> PySersicSourcePrior:
        """Function to generate default priors based on a given image and profile type

        Parameters
        ----------
        profile_type : str
            Type of profile
        sky_type : str, default 'none'
            Type of sky model to use, must be one of: 'none', 'flat', 'tilted-plane'
        suffix : str, default ''
            Add suffix onto all source related variables, generally only used to number sources in MultiPrior
        Returns
        -------
        PySersicSourcePrior
            Pysersic prior object to be used to initialize FitSingle
        """
        
        prior = PySersicSourcePrior(profile_type=profile_type, sky_type= sky_type,sky_guess=self.sky_guess,sky_guess_err=self.sky_guess_err, suffix=suffix)

        # 3 properties common to all sources
        prior.set_gaussian_prior('flux',self.flux_guess,self.flux_guess_err)
        prior.set_gaussian_prior('xc', self.xc_guess, 1.)
        prior.set_gaussian_prior('yc', self.yc_guess, 1.)

        if profile_type in ['exp','dev','sersic']:
            prior.set_truncated_gaussian_prior('r_eff', self.r_eff_guess,self.r_eff_guess_err, low = 0.5)
            prior.set_uniform_prior('ellip', 0, 0.9)
            prior.set_custom_prior('theta', dist.VonMises(loc = self.theta_guess,concentration=2), reparam= infer.reparam.CircularReparam() )
            if profile_type == 'sersic':
                prior.set_uniform_prior('n', 0.65, 8)

        elif profile_type == 'doublesersic':

            prior.set_uniform_prior('f_1', 0.,1.)

            prior.set_custom_prior('theta', dist.VonMises(loc = self.theta_guess,concentration=2), reparam= infer.reparam.CircularReparam() )

            r_loc1 = self.r_eff_guess/1.5
            r_eff_guess_err1 = jnp.sqrt(self.r_eff_guess/1.5)
            prior.set_truncated_gaussian_prior('r_eff_1', r_loc1,r_eff_guess_err1, low = 0.5)

            r_loc2 = self.r_eff_guess*1.5
            r_eff_guess_err2 = jnp.sqrt(self.r_eff_guess*1.5)
            prior.set_truncated_gaussian_prior('r_eff_2', r_loc2,r_eff_guess_err2, low = 0.5)


            prior.set_uniform_prior('ellip_1', 0,0.9)
            prior.set_uniform_prior('ellip_2', 0,0.9)

            prior.set_truncated_gaussian_prior('n_1',4,1, low = 0.65,high = 8)
            prior.set_truncated_gaussian_prior('n_2',1,1, low = 0.65,high = 8)

        
        return prior
    
    def visualize(self,figsize: Tuple[float,float]= (6.,6.),cmap:str = 'gray',scale: float = 1.)-> None:
        """Display a figure summarizing the current guess for the source properties

        Parameters
        ----------
        figsize : Tuple[float,float], optional
            figure, by default (6.,6.)
        cmap : str, optional
            color map, by default 'gray'
        scale : float, optional
            number of +/- std's at which to clip image, by default 1
        """
        if not hasattr(self,'flux_guess'):
            self.set_flux_guess() 
        if not hasattr(self,'r_eff_guess'):
            self.set_r_eff_guess()
        if not hasattr(self,'theta_guess'):
            self.set_theta_guess() 
        if not hasattr(self,'xc_guess'):
            self.set_position_guess()
        
        fig, ax = plt.subplots(figsize=figsize,)
        if self.mask is not None:
            image= np.ma.masked_array(self.image,mask=self.mask)
        else:
            image = self.image
        vmin = np.mean(image) - scale*np.std(image)
        vmax = np.mean(image) + scale*np.std(image)
        ax.imshow(image,cmap=cmap,origin='lower',vmin=vmin,vmax=vmax)
        ax.plot(self.xc_guess,self.yc_guess,'x',color='r')
        
        dx = 10*self.r_eff_guess*np.cos(self.theta_guess+np.pi/2.)
        dy = 10*self.r_eff_guess*np.sin(self.theta_guess+np.pi/2.)
        ax.arrow(self.xc_guess,self.yc_guess,dx=dx,dy=dy,width=0.1,head_width=1)
        arr = np.linspace(0,2*np.pi,100)
        x = np.cos(arr) * self.r_eff_guess + self.xc_guess 
        y = np.sin(arr) * self.r_eff_guess + self.yc_guess 
        ax.plot(x,y,'r',lw=2)
        plt.show()

def estimate_sky(image: np.array, mask: Optional[np.array] = None, n_pix_sample:int = 5 )-> Tuple[float,float,int]:
    """Estimate the sky background using the edge of the cutout

    Parameters
    ----------
    im : np.array
        image, either an array or masked array
    mask : Optional[np.array], optional
        mask to apply, if im is not a masked array already, by default None
    n_pix_sample : int, optional
        number of pixels around the edge to use, by default 5

    Returns
    -------
    Tuple[float,float,int]
        a tuple containing the median, standard deviation and number of pixels used
    """
    if not np.ma.is_masked(image) and not mask is None:
        image = np.ma.masked_array(image, mask )
    edge_pixels = np.concatenate((image[:n_pix_sample,:],image[-n_pix_sample:,:],image[n_pix_sample:-n_pix_sample,:n_pix_sample],image[n_pix_sample:-n_pix_sample,-n_pix_sample:]),axis=None)
    median_val = np.ma.median(edge_pixels)
    err_on_median = bws(edge_pixels)
    return median_val, err_on_median, np.prod(edge_pixels.shape)

def autoprior(image: np.array, profile_type: 'str', mask:np.array =None, sky_type: str = 'none') -> PySersicSourcePrior:
    """Simple wrapper function to generate a prior using the built-in defaults. This can be used as a starting place but may not work for all sources

    Parameters
    ----------
    image : np.array
        science image
    profile_type : str
        Type of profile to be fit
    mask : np.array, optional
        pixel by pixel mask, by default None
    sky_type : str
        Type of sky to fit, default 'none'

    Returns
    -------
    PySersicSourcePrior
        Prior object that can be used in initializing FitSingle
    """
    props = SourceProperties(image = image, mask = mask)
    prior = props.generate_prior(profile_type = profile_type, sky_type = sky_type)
    return prior




from pysersic.rendering import PixelRenderer,FourierRenderer,HybridRenderer, BaseRenderer
from pysersic.rendering import render_sersic_2d
from astropy.convolution import Gaussian2DKernel
import pytest
import jax.numpy as jnp
from jax.scipy.special import gammainc
from scipy.integrate import dblquad
from functools import partial
 
kern = Gaussian2DKernel(x_stddev= 1.5)
psf = kern.array
err_tol = 0.01 # 1% error tolerence for total flux

@pytest.mark.parametrize("renderer", [PixelRenderer,FourierRenderer,HybridRenderer])
@pytest.mark.parametrize("pos", [(50,50),(50.5,50.5),(50.25,50.25) ,(50.,50.5),(50.5,50.)]) 
def test_point_source(renderer,pos):
    flux = 10.
    renderer_test = renderer((100,100), psf)
    im = renderer_test.render_pointsource(pos[0],pos[1],flux)
    assert pytest.approx(im.sum(), rel = err_tol) == flux


@pytest.mark.parametrize("renderer", [PixelRenderer,FourierRenderer,HybridRenderer])
@pytest.mark.parametrize("pos", [(75.,75.),(75.5,75.5),])
@pytest.mark.parametrize("re", [3.,5.]) 
@pytest.mark.parametrize("n", [1.5,2.5]) 
@pytest.mark.parametrize("ellip", [0,0.5])
@pytest.mark.parametrize("theta", [0,3.14/4.]) 
def test_sersic(renderer,pos,re,n,ellip,theta):
    renderer_test = renderer((150,150), psf)
    flux = 10
    #Calculate fraction of flux contained in image
    int_test = partial(render_sersic_2d, xc = pos[0],yc = pos[1], flux = 10, r_eff = re, n = n, ellip = ellip,theta = theta)
    to_int = lambda x,y: float(int_test(x,y))
    lo_fun = lambda x: 0.
    hi_fun = lambda x: 150. 
    flux_int,_ = dblquad(to_int, 0.,150., lo_fun,hi_fun,epsrel=5.e-3)
    
    im = renderer_test.render_sersic(pos[0],pos[1],flux,re,n,ellip,theta)
    total = float( jnp.sum(im) )
    assert pytest.approx(flux_int, rel = err_tol) == total

@pytest.mark.parametrize("renderer", [PixelRenderer,FourierRenderer,HybridRenderer])
@pytest.mark.parametrize("prof", ['exp','dev']) 
@pytest.mark.parametrize("pos", [(75.,75.),(75.5,75.5),])
@pytest.mark.parametrize("re", [3.,5.]) 
@pytest.mark.parametrize("ellip", [0,0.5])
@pytest.mark.parametrize("theta", [0,3.14/4.]) 
def test_exp_dev(renderer,prof,pos,re,ellip,theta):
    flux = 10.
    renderer_test = renderer((150,150), psf)
    if prof == 'exp':
        im = renderer_test.render_exp(pos[0],pos[1],flux,re,ellip,theta)
        n=1.
    if prof == 'dev':
        im = renderer_test.render_dev(pos[0],pos[1],flux,re,ellip,theta)
        n=4.

    #Calculate fraction of flux contained in image
    int_test = partial(render_sersic_2d, xc = pos[0],yc = pos[1], flux = 10, r_eff = re, n = n, ellip = ellip,theta = theta)
    to_int = lambda x,y: float(int_test(x,y))

    lo_fun = lambda x: 0.
    hi_fun = lambda x: 150. 
    flux_int,_ = dblquad(to_int, 0.,150., lo_fun,hi_fun,epsrel=5.e-3)

    total = float( jnp.sum(im) )
    assert pytest.approx(flux_int, rel = err_tol) == total


def test_sky_rendering():
    renderer_test = BaseRenderer((100,100), psf)

    assert renderer_test.render_sky(0, None) == 0

    assert renderer_test.render_sky(1., 'flat') == 1

    sky_1 = renderer_test.render_sky([1e-4,0.,0.], 'tilted-plane')
    assert sky_1.shape == (100,100)
    sky_sum_1 = float(jnp.sum(sky_1))
    assert pytest.approx(1e-4*100*100, rel = 1e-6) == sky_sum_1

    sky_2 = renderer_test.render_sky([0.,1e-3,0.], 'tilted-plane')
    assert sky_2.shape == (100,100)
    sky_sum_2 = float(jnp.sum(sky_2))
    assert pytest.approx(0, abs = 1e-4) == sky_sum_2

    sky_3 = renderer_test.render_sky([0.,0.,1e-3,], 'tilted-plane')
    assert sky_3.shape == (100,100)
    sky_sum_3 = float(jnp.sum(sky_3))
    assert pytest.approx(0, abs = 1e-4) == sky_sum_3


import pytest
import jax.numpy as jnp
from pysersic import loss
from numpyro.handlers import seed

loss_func_list = [loss.gaussian_loss, loss.cash_loss,loss.gaussian_loss_w_frac,
                loss.gaussian_loss_w_sys, loss.student_t_loss, loss.student_t_loss_free_sys,
                loss.pseudo_huber_loss, loss.gaussian_mixture, loss.gaussian_mixture_w_sys, loss.gaussian_mixture_w_frac
]


data = jnp.ones((100,100))
sig = jnp.ones((100,100))
rms = jnp.ones((100,100))
mask = jnp.ones((100,100), dtype=bool)

@pytest.mark.parametrize("loss_func", loss_func_list)
def test_all_losses(loss_func):
    with seed(rng_seed=0):
        loss_val = loss_func(data,sig,rms,mask)


from typing import Iterable, Union

import jax.numpy as jnp
from astropy.io import fits

from pysersic import FitMulti, FitSingle, priors

from .loss import gaussian_loss


def search_start(list_of_strings: Iterable, to_match: str)-> list:
    """Search a list of strings and match those which start with to_match

    Parameters
    ----------
    list_of_strings : Iterable
    to_match : str
        pattern to match

    Returns
    -------
    list
        List of results contianing the index and full string
    """
    result = []
    for j,string in enumerate(list_of_strings):
        if string.split()[0] == to_match:
            result.append([j,string])
    return result

def load_fits_from_galfit_line(line:str) -> jnp.array:
    """Load fits file from galfit config line

    Parameters
    ----------
    line : str

    Returns
    -------
    jnp.array
        Image in first HDU
    """
    fname = line.split()[1]
    return jnp.array( fits.getdata(fname) )

def match_line_and_load_galfit(list_of_strings: Iterable, to_match: str, assert_message: str = "")-> jnp.array:
    """Combine the above two functions to search and load fits files from a galfit config

    Parameters
    ----------
    list_of_strings : Iterable
    to_match : str
    assert_message : str, optional
        Message to raise if line file is not present

    Returns
    -------
    jnp.array
        _description_

    Raises
    ------
    UserWarning
        _description_
    """
    matches = search_start(list_of_strings, to_match)
    line = matches[0][1]
    if line.split()[1] == 'none':
        raise UserWarning(assert_message)
    return load_fits_from_galfit_line(line)

def generate_fitter_from_galfit_config(config_loc: str, loss: callable = gaussian_loss ) -> Union[FitSingle,FitMulti]:
    """Function to generate a Fitter instance from a galfit config file. This function is only meant to be a starting place to easily try PySersic on the many galfit config files we know you all have lying around. As such this does not follow all of the rules and constraints the galfit takes into account. For more fine grained control we recommend initializing a pysersic Fitter instance yourself, please see the `examples/` folder or the documentation for some help with this. 

    Parameters
    ----------
    config_loc : str
        Config file locations
    loss : callable, optional
        Loss function to use, see loss.py for more details, by default gaussian_loss

    Returns
    -------
    Union[FitSingle,FitMulti]
        Fitter based on config file. If there is only one source then returns a FitSingle object, otherwise returns a fit_multi
    """
    
    with open(config_loc,'r') as f:
        lines = []
        for line in f.readlines():
            if line[0] == '#' or line[:1] == '\n' or line[0] == '=':
                continue
            else:
                lines.append(line)
    
    im = match_line_and_load_galfit(lines, 'A)', assert_message='Missing science image, labelled D) in config file')
    rms = match_line_and_load_galfit(lines, 'C)', assert_message='Missing rms image, labelled C) in config file')
    psf = match_line_and_load_galfit(lines,'D)', assert_message='Missing psf image, labelled D) in config file')

    try:
        mask = match_line_and_load_galfit(lines,'F)')
    except:
        mask = jnp.zeros_like(im)
    zpt = float ( search_start(lines, 'J)')[0][1].split()[1] )
    source_list = search_start(lines, '0)')
    
    sky_type = 'none'
    source_dict = {}
    source_dict['type'] = []
    source_dict['flux'] = []
    source_dict['r'] = []
    source_dict['x'] = []
    source_dict['y'] = []

    for i,source in enumerate(source_list):
        ind = source[0]
        
        if i == len(source_list) -1:
            next_ind = None
        else:
            next_ind = source_list[i+1][0]
        source_string_list = lines[ind:next_ind]

        source_type = source[1].split()[1]
        if source_type == 'sersic' or 'dev' in source_type or 'exp' in source_type or source_type == 'psf':
            if 'dev' in source_type:
                source_type = 'dev'
            elif 'exp' in source_type:
                source_type = 'exp'
            elif source_type == 'psf':
                source_type = 'pointsource'

            mag = float( search_start(source_string_list, '3)')[0][1].split()[1] )
            flux = 10.**( (mag - zpt)/-2.5)
            
            x,y = search_start(source_string_list, '1)')[0][1].split()[1:3]
            x = float(x)
            y = float(y)
    
            source_dict['type'].append(source_type)
            source_dict['x'].append(x)
            source_dict['y'].append(y)
            source_dict['flux'].append(flux)

            if source_type != 'pointsource':
                r = float( search_start(source_string_list, '4)')[0][1].split()[1] )
            else:
                r = -99
            source_dict['r'].append(r)

        elif source_type == 'sky':
            print('sky')
            fit_back = search_start(source_string_list, '1)')[0][1].split()[2]
            back_init = float(search_start(source_string_list, '1)')[0][1].split()[1])
            fit_x_sl = search_start(source_string_list,'2)')[0][1].split()[2]
            fit_y_sl = search_start(source_string_list,'2)')[0][1].split()[2]
            if fit_y_sl == '1' or fit_x_sl == '1':
                sky_type = 'tilted-plane'
            elif fit_back == '1':
                sky_type = 'flat'
            else:
                sky_type = 'none'
        else:
            print (f"Skipping source {i}, profile type of {source_type} currently not supported")

    N_sources = len(source_dict['type'])
    types = source_dict['type']
    print(f"Found {N_sources} source(s) to fit of type(s) {types}")

    if len(source_dict['type']) == 1:
        prof_type = source_dict['type'][0]
        sp = priors.SourceProperties(im, mask = mask)
        sp.set_sky_guess(sky_guess = 0, sky_guess_err = 1)
        sp.set_flux_guess(flux_guess= source_dict['flux'][0])
        sp.set_position_guess( [source_dict['x'][0],source_dict['y'][0] ])

        if prof_type != 'psf':
            sp.set_r_eff_guess(r_eff_guess=source_dict['r'][0])
        prior = sp.generate_prior(prof_type, sky_type = sky_type)
        ps_fitter = FitSingle(data = im, rms = rms, psf = psf, prior = prior, loss_func = loss )

    elif len(source_dict['type']) > 1:
        multi_prior = priors.PySersicMultiPrior(catalog=source_dict, sky_type = sky_type,sky_guess=back_init, sky_guess_err = back_init + 1e-5)
        ps_fitter = FitMulti(data = im, rms = rms, psf = psf, prior = multi_prior, loss_func = loss)

    else:
        print ('Did not find any viable sources in galfit config')
        ps_fitter = 0
    
    return ps_fitter



import setuptools

setuptools.setup(
    name="pysersic",
    version="0.1.2",
    author="Imad Pasha & Tim Miller",
    author_email="imad.pasha@yale.edu",
    description="A Tool for fitting sersic profiles in python",
    download_url="https://github.com/pysersic/pysersic/archive/refs/tags/v0.1.2.tar.gz",
    packages=["pysersic",],
    install_requires=['numpy',
        'scipy',
        'matplotlib',
        'pandas',
        'astropy',
        'corner',
        'photutils>=0.16',
        'jax',
        'numpyro',
        'arviz',
        'tqdm',
        'asdf',
        ]
 )



class RMSWarning(Warning):
    pass
class KernelError(Exception):
    pass
class PSFNormalizationWarning(Warning):
    pass
class MaskWarning(Warning):
    pass

class ShapeMatchError(Exception):
    pass